#!/usr/bin/env python3
# system modules
import logging
import http.client
import urllib
import json
import contextlib

# internal modules
from .. import utils
from . import responsetypes
from .variables import *
from .errors import *


class ApiRequest(object):
    """ Class for api requests
    """
    def __init__(self, server, url, payload = {}):
        """ class constructor
        Args:
            server (str): the server domain
            url (str): the url relative to the server
            payload [Optional(dict)]: the payload
        """
        self.server = server
        self.url = url
        self.payload = payload

    ##################
    ### Properties ###
    ##################
    @property
    def logger(self):
        """ the logging.Logger used for logging.
        Defaults to logging.getLogger(__name__).
        """
        try: # try to return the internal property
            return self._logger 
        except AttributeError: # didn't work
            return logging.getLogger(__name__) # return default logger
    
    @logger.setter
    def logger(self, logger): # pragma: no cover
        assert isinstance(logger, logging.Logger), \
            "logger property has to be a logging.Logger"
        self._logger = logger
    
    @property
    def api_connection(self):
        """ The http.client.HTTPSConnection used for server communication.
        You may set this, otherwise a new instance is created automatically on
        first use.
        """
        try: # try to return the internal attribute
            return self._api_connection
        except AttributeError: # didn't work
            # set it to new instance
            self._api_connection = http.client.HTTPSConnection(
                NETATMO_API_SERVER)
        # return internal attribute
        return self._api_connection

    @api_connection.setter
    def api_connection(self, newconn): # pragma: no cover
        if not isinstance(newconn, http.client.HTTPSConnection):
            self.logger.debug("authentication property needs to be of " 
                "class HTTPSConnection. Using empty instance instead " 
                "of {}.".format(newauth))
            self._api_connection = http.client.HTTPSConnection(
                NETATMO_API_SERVER)
        else:
            self._api_connection = newconn

    @property
    def server(self):
        """ the api server domain
        """
        try: return self._server
        except AttributeError: return "" # pragma: no cover

    @server.setter
    def server(self, newserver):
        self._server = str(newserver)

    @property
    def url(self):
        """ the url relative to the server for this api request
        """
        try: return self._url
        except AttributeError: return "" # pragma: no cover

    @url.setter
    def url(self, newurl):
        self._url = str(newurl)

    @property
    def payload(self):
        """ the payload to send to the api
        """
        try: return self._payload
        except AttributeError: return {} # pragma: no cover
        
    @payload.setter
    def payload(self, newpayload):
        try:
            assert isinstance(newpayload, dict)
            json.dumps(newpayload)
            urllib.parse.urlencode(newpayload)
        except (TypeError,AssertionError): # pragma: no cover
            raise InvalidPayloadError("payload has to be a simple dict")
        self._payload = newpayload

    @property
    def payload_urlencoded(self):
        """ [read only] return the urlencoded payload (UTF-8)
        """
        return urllib.parse.urlencode(self.payload, encoding="UTF-8")

    @property
    def response(self):
        """ the api response to this request. Instance of ApiResponse or
        derivates. Issues the request if necessary. You should not set this
        property.  
        """
        try:
            return self._response
        except AttributeError:
            self.issue() # issue the request
        return self._response # now return

    @response.setter
    def response(self, newresponse):
        try:
            assert issubclass(newresponse.__class__, responsetypes.ApiResponse)
        except (AssertionError,AttributeError): # pragma: no cover
            raise TypeError("response property has to be instance of " 
                "ApiResponse or derivates.")
        self._response = newresponse

    ########################
    ### context managers ###
    ########################
    @contextlib.contextmanager
    def connection_to_api(self):
        self.connect_to_api()
        try:
            yield
        finally:
            self.close_api_connection()

    ###############
    ### methods ###
    ###############
    def connect_to_api(self):
        """ Connect to the api server 
        """
        self.logger.debug("connecting to api server...")
        self.api_connection.connect() # connect to server
        self.logger.debug("connected to api server.")

    def close_api_connection(self):
        """ Disconnect from the api server 
        """
        self.logger.debug("closing connection to api server...")
        self.api_connection.close() # close the connection
        self.logger.debug("connection to api server closed")

    def post_request(self):
        """ Issue a POST request to the api server on the given url with the
        specified payload.
        Returns:
            response (dict): JSON decoded response data
        """
        with self.connection_to_api(): # connect to api
            self.logger.debug("issuing post request to {u} on server {s} with " 
                "payload {p}".format(s=self.server,u=self.url,
                p=self.payload_urlencoded))
            # POST request to oauth/token
            self.api_connection.request(
                method  = "POST",                  # a POST request
                url     = self.url,                # to this relative url
                body    = self.payload_urlencoded, # with this payload
                headers = PLAIN_URLENCODED_HEADERS # and this header
                )

            # evaluate output
            response  = self.api_connection.getresponse() # get the responsse
            data      = response.read().decode()          # read and decode
            data_json = json.loads(data)                  # load the JSON

            self.logger.debug("api responeded: {}".format(data_json))

        return data_json # return the json data
        
    def issue(self): # pragma: no cover
        """ Issue a POST request to the api server on the given url with the
        specified payload and set the 'response' property to an ApiResponse
        object. Subclasses may override this class to use specific derivates
        of ApiResponse.
        """
        # post the request
        res_data = self.post_request()
        # pack the response into an ApiResponse object
        response = responsetypes.ApiResponse(
            response = res_data, request = self)
        # set the response property
        self.response = response

    def __repr__(self): # pragma: no cover
        """ python representation of this object
        """
        # self.logger.debug("__repr__ called")
        reprstring = ("{classname}(\n" 
              "server = {server},\n"
              "url = {url},\n"
              "payload = {payload},\n"
              ")").format(
            classname="{module}.{name}".format(
                name=self.__class__.__name__,module=self.__class__.__module__),
            server = self.server.__repr__(),
            url = self.url.__repr__(),
            payload = self.payload.__repr__(),
            )
        return reprstring


class TokenRequest(ApiRequest):
    """ base class for token requests
    """
    def __init__(self, payload = {}):
        """ class constructor
        Args:
            payload [Optional(dict)]: the payload
        """
        super().__init__( 
            server = NETATMO_API_SERVER,
            url = NETATMO_API_TOKEN_URL,
            payload = payload
            )

    def issue(self):
        """ Issue a POST request to the api server on the Oauth2 endpoint
        with the specified payload and set the 'response' property to an
        TokenResponse object.
        """
        # post the request
        res_data = self.post_request()
        # pack the response into an ApiResponse object
        response = responsetypes.TokenResponse(response = res_data,
            request = self)
        # set the response property
        self.response = response

class GetpublicdataRequest(ApiRequest):
    """ class for Getpublicdata requests
    """
    def __init__(self, payload = {}):
        """ class constructor
        Args:
            payload [Optional(dict)]: the payload
        """
        super().__init__( 
            server = NETATMO_API_SERVER,
            url = NETATMO_API_GETPUBLICDATA_URL,
            payload = payload
            )

    def issue(self):
        """ Issue a POST request to the api server on the Getpublicdata endpoint
        with the specified payload and set the 'response' property to an
        GetpublicdataResponse object.
        """
        # post the request
        res_data = self.post_request()
        # pack the response into an ApiResponse object
        response = responsetypes.GetpublicdataResponse(response = res_data,
            request = self)
        # set the response property
        self.response = response
        

class GetmeasureRequest(ApiRequest):
    """ class for Getmeasure requests
    """
    def __init__(self, payload = {}):
        """ class constructor
        Args:
            payload [Optional(dict)]: the payload
        """
        super().__init__( 
            server = NETATMO_API_SERVER,
            url = NETATMO_API_GETMEASURE_URL,
            payload = payload
            )

    def issue(self):
        """ Issue a POST request to the api server on the Getmeasure endpoint
        with the specified payload and set the 'response' property to an
        GetmeasureResponse object.
        """
        # post the request
        res_data = self.post_request()
        # pack the response into an ApiResponse object
        response = responsetypes.GetmeasureResponse(response = res_data,
            request = self)
        # set the response property
        self.response = response



class GetstationsdataRequest(ApiRequest):
    """ class for Getstationsdata requests
    """
    def __init__(self, payload = {}):
        """ class constructor
        Args:
            payload [Optional(dict)]: the payload
        """
        super().__init__( 
            server = NETATMO_API_SERVER,
            url = NETATMO_API_GETSTATIONSDATA_URL,
            payload = payload
            )

    def issue(self):
        """ Issue a POST request to the api server on the Getstationsdata 
        endpoint with the specified payload and set the 'response' property to
        an GetstationsdataResponse object.
        """
        # post the request
        res_data = self.post_request()
        # pack the response into an ApiResponse object
        response = responsetypes.GetstationsdataResponse(response = res_data,
            request = self)
        # set the response property
        self.response = response



#!/usr/bin/env python3
# system modules
import logging
import time

# internal modules
from .. import utils
from . import authentication
from . import responsetypes
from . import requests
from .variables import *
from .errors import *


class NetatmoClient(object):
    """
    Netatmo api client
    
    Args:
        authentication (Authentication, optional): 
            the authentication used for Oauth2 authentication
    """
    def __init__(self, authentication = None):
        self.authentication = authentication


    ##################
    ### Properties ###
    ##################
    @property
    def logger(self):
        """ the logging.Logger used for logging.
        Defaults to logging.getLogger(__name__).
        """
        try: # try to return the internal property
            return self._logger 
        except AttributeError: # didn't work
            return logging.getLogger(__name__) # return default logger
    
    @logger.setter
    def logger(self, logger): # pragma: no cover
        assert isinstance(logger, logging.Logger), \
            "logger property has to be a logging.Logger"
        self._logger = logger
    
    @property
    def authentication(self): # pragma: no cover
        """ The Authentication used for server authentication
        Defaults to an empty instance of Authentication on first use.
        """
        try: # try to return the internal attribute
            return self._authentication
        except AttributeError: # didn't work
            # set it to new instance
            self._authentication = authentication.Authentication()
        # return internal attribute
        return self._authentication

    @authentication.setter
    def authentication(self, newauth):
        if not isinstance(newauth, authentication.Authentication):
            self.logger.debug("authentication property needs to be of " 
                "class Authentication. Using empty instance instead " 
                "of {}.".format(newauth))

            self._authentication = authentication.Authentication()
        else:
            self._authentication = newauth


    ###############
    ### Methods ###
    ###############
    def Getpublicdata(self,region,required_data=None,filter=False):
        """ Issue a Getpublicdata POST request to the netatmo server

        Args:
            region (dict): dict definig the desired request region. Required
                keys: lat_ne [-85;85], lat_sw [-85;85], lon_ne [-180;180] and
                lon_sw [-180;180] with lat_ne > lat_sw and lon_ne > lon_sw
            required_data (str or None, optional): Defaults to None which means
                no filter.
            filter (bool, optional): server-side filter for stations with
                unusual data. Defaults to False with means no filter

        Returns:
            instance of GetpublicdataResponse with response data

        Raises:
            InvalidApiInputError or derivates if wrong input was provided
            ApiResponseError or derivates if api responded with error

        Returns:
            GetpublicdataResponse: The api response
        """

        ### Check the input ###
        # check the region
        for key,bounds in GETPUBLICDATA_REGION_BOUNDS.items():
            try:
                if region[key] < bounds[0] or region[key] > bounds[1]:
                    raise InvalidRegionError
            except:
                raise InvalidRegionError
        if region["lat_ne"] < region["lat_sw"] \
            or region["lon_ne"] < region["lon_sw"]:
                raise InvalidRegionError
        # check the required_data option
        if required_data not in GETPUBLICDATA_ALLOWED_REQUIRED_DATA \
            and required_data is not None:
            raise InvalidRequiredDataError
        # check the filter option
        if not isinstance(filter,bool):
            raise InvalidApiInputError("'filter' needs to be bool")

            
        ### Create the payload ###
        payload = {} # start with empty dict
        # set the access token
        payload["access_token"] = self.authentication.tokens.get("access_token")
        # set the region
        payload.update(region)
        # set the required_data
        if required_data is not None: 
            payload["required_data"] = str(required_data)
        # set the filter option
        payload["filter"] = str(filter).lower()

        ### issue the request ###
        apirequest = requests.GetpublicdataRequest( payload = payload )
        apiresponse = apirequest.response

        ### check for errors ###
        error = apiresponse.response.get("error",{}).get("message")
        if error: raise API_ERRORS.get(error,ApiResponseError(error))

        return apiresponse


    def Getmeasure(self,device_id,module_id=None,type=None,
        scale=None,date_begin=None,date_end=None,real_time=False,
        optimize=False):
        """ Issue a Getmeasure POST request to the netatmo server
        Taken from API documentation: 
        https://dev.netatmo.com/dev/resources/technical/reference/common/getmeasure

        Args:
            device_id (str): The mac address of the device
            module_id (str, optinal): The mac address of the module of
                interest. If not specified, returns data of the device. If
                specified, returns data from the specified module
            type (list of str, optional): Measures interested in. List of
                "rain","humidity","pressure","wind" and "temperature".
            scale (str, optional): Timelapse between two measurements. "max"
                (every value is returned), "30min" (1 value every 30min),
                "1hour", "3hours", "1day", "1week", "1month". Defaults to "max".
            date_begin,date_end (int, optional): UNIX-timestamp of first/last
                measure to receive. Limit is 1024 measures.
            optimize (bool, optional): Determines the format of the answer.
                Default is False. For mobile apps we recommend True and False if
                bandwidth isn't an issue as it is easier to parse.
            real_time (bool, optional): If scale different than max, timestamps
                are by default offset + scale/2. To get exact timestamps, use
                true. Default is false.  

        Returns:
            GetmeasureResponse: The api response
        """
        ### Check the input ###
        # check device_id
        if not MAC_ADDRESS_REGEX.match(device_id):
            raise InvalidApiInputError("device_id = '{}' doesn't look like " 
                "MAC address".format(device_id))
        # check module_id
        if not module_id is None:
            if not MAC_ADDRESS_REGEX.match(module_id):
                raise InvalidApiInputError("module_id = '{}' doesn't look like " 
                    "MAC address".format(module_id))
        # check scale
        if not scale is None: # specified
            if not scale in GETMEASURE_ALLOWED_TYPES_BY_SCALE.keys():
                raise InvalidApiInputError("scale must be one of {}".format(
                    list(GETMEASURE_ALLOWED_TYPES_BY_SCALE.keys())))
        else: # not specified
            scale = "max" # use all values
        # check type
        allowed_types = GETMEASURE_ALLOWED_TYPES_BY_SCALE.get(scale,[])
        if not type is None: # specified
            if not all([x in allowed_types for x in type]):
                raise InvalidApiInputError("at given scale '{}' type must be a " 
                    "sublist of {}".format(scale,allowed_types))
        else: # not specified
            type = allowed_types # use all
        # check date_begin
        if not date_begin is None: 
            try: 
                date_begin = int(date_begin)
                if date_begin < 0: raise Exception
                if date_begin > time.time():
                    self.logger.warning(
                        "Specified date_begin is in the future!")
            except:
                raise InvalidApiInputError("date_begin must be UNIX-timestamp")
        if not date_end is None: 
            try: 
                date_end = int(date_end)
                if date_end < 0: raise Exception
            except:
                raise InvalidApiInputError("date_end must be UNIX-timestamp")
        try: 
            if date_end < date_begin: # compare begin and end
                raise InvalidApiInputError("date_end must be greater than " 
                    "date_begin")
        except TypeError: pass
        # check the optimize option
        if not isinstance(optimize,bool):
            raise InvalidApiInputError("optimize must be bool")
        # check real_time option
        if not isinstance(real_time,bool):
            raise InvalidApiInputError("real_time must be bool")


        ### Prepare the payload ###
        payload = {} # start with empty dict
        # access token
        payload["access_token"] = self.authentication.tokens.get("access_token")
        # device id
        payload["device_id"] = device_id
        # module id
        if not module_id is None:
            payload["module_id"] = module_id
        # scale
        payload["scale"] = scale
        # type
        payload["type"] = ",".join(type)
        # date_begin
        if not date_begin is None:
            payload["date_begin"] = date_begin
        # date_end
        if not date_end is None:
            payload["date_end"] = date_end
        # optimize
        payload["optimize"] = str(optimize).lower()
        # real_time
        payload["real_time"] = str(real_time).lower()

        ### issue the request ###
        apirequest = requests.GetmeasureRequest( payload = payload )
        apiresponse = apirequest.response

        ### check for errors ###
        error = apiresponse.response.get("error",{}).get("message")
        if error: raise API_ERRORS.get(error,ApiResponseError(error))

        return apiresponse


    def Getstationsdata(self, device_id, get_favourites = False):
        """ Issue a Getstationsdata POST request to the netatmo server
        Taken from API documentation: 

        Args:
            device_id (str): The mac address of the device
            get_favourites (bool): To retrieve user's favorite weather stations.
                Is converted to bool. Default is false.

        Returns:
            GetstationsdataResponse: The api response
        """
        ### Check the input ###
        # check device_id
        if not MAC_ADDRESS_REGEX.match(device_id):
            raise InvalidApiInputError("device_id = '{}' doesn't look like " 
                "MAC address".format(device_id))
        # check get_favourites
        try: get_favourites = bool(get_favourites) # try to convert to bool
        except:
            raise InvalidApiInputError("get_favourites has to be bool or at " 
                "least convertible to bool")

        ### create the payload ###
        payload = {} # start with empty dict
        # access token
        payload["access_token"] = self.authentication.tokens.get("access_token")
        # device id
        payload["device_id"] = device_id
        # get_favourites
        payload["get_favourites"] = str(get_favourites).lower()
        
        ### issue the request ###
        apirequest = requests.GetstationsdataRequest( payload = payload )
        apiresponse = apirequest.response

        ### check for errors ###
        error = apiresponse.response.get("error",{}).get("message")
        if error: raise API_ERRORS.get(error,ApiResponseError(error))

        return apiresponse


    def __repr__(self): # pragma: no cover
        """ python representation of this object
        """
        # self.logger.debug("__repr__ called")
        reprstring = ("{classname}(\n" 
              "authentication = {authentication}\n"
              ")").format(
            classname="{module}.{name}".format(
                name=self.__class__.__name__,module=self.__class__.__module__),
            authentication = self.authentication.__repr__()
            )
        return reprstring






#!/usr/bin/env python3
# system modules
import logging

# external modules
import pandas as pd
from pandas import DataFrame
import numpy as np

# internal modules
from .errors import *
from . import requests

class ApiResponse(object):
    """ Base class for Netatmo api response datasets
    """
    def __init__(self, request, response):
        """ Class constructor
        Args:
            request (instance of ApiRequest or derivate): the api request
            response (dict): The raw api response
        """
        self.response = response
        self.request = request

    ##################
    ### Properties ###
    ##################
    @property
    def logger(self): # pragma: no cover
        """ the logging.Logger used for logging.
        Defaults to logging.getLogger(__name__).
        """
        try: # try to return the internal property
            return self._logger 
        except AttributeError: # didn't work
            return logging.getLogger(__name__) # return default logger
    
    @logger.setter
    def logger(self, logger): # pragma: no cover
        assert isinstance(logger, logging.Logger), \
            "logger property has to be a logging.Logger"
        self._logger = logger
    
    @property
    def response(self):
        """ The raw api response dict.
        """
        try: # try to return internal attribute
            return self._response
        except AttributeError: # pragma: no cover
            self._response = {}
        return self._response # return internal attribute

    @response.setter
    def response(self, newresponse):
        assert isinstance(newresponse, dict), \
            "reponse property has to be of class dict"
        self._response = newresponse

    @property
    def request(self):
        """ The raw api request dict.
        """
        return self._request # return internal attribute

    @request.setter
    def request(self, newrequest):
        assert issubclass(newrequest.__class__, requests.ApiRequest), \
            "reponse property has to be of class ApiRequest or derivates"
        self._request = newrequest
    
    def __repr__(self): # pragma: no cover
        """ python representation of this object
        """
        # self.logger.debug("__repr__ called")
        reprstring = ("{classname}(\n" 
              "response = {response},\n"
              "request = {request},\n"
              ")").format(
            classname="{module}.{name}".format(
                name=self.__class__.__name__,module=self.__class__.__module__),
            # compact version
            response = self.response.__repr__(),
            # pretty version
            # response = json.dumps(self.response,sort_keys=True,indent=4)
            # the request
            request = self.request.__repr__()
            )
        return reprstring


class TokenResponse(ApiResponse):
    """ Class that holds the responded data of a Oauth2 token request
    """
    pass


class GetpublicdataResponse(ApiResponse):
    """ Class that holds the responded data of a Getstationdata request
    """
    def dataframe(self, only_inside=False):
        """ 
        Convert the response to a pandas.DataFrame

        Args:
            only_inside (bool, optional): Drop stations outside the requested
                region? Defaults to ``False``.

        Returns:
            pandas.DataFrame: The response converted to a DataFrame
        """
        # get the list of stations
        stations = self.response.get("body")
        if not isinstance(stations,list): # pragma: no cover
            raise ApiResponseError("'body' part of response does not " 
                "exist or is no list.")

        # start with basic dict
        stationdict = { 
            "id": [],
            "altitude": [],
            "longitude": [],
            "latitude": [],
            "timezone": [],
            }
        # loop over all stations
        for station in stations:
            station_df = DataFrame() # start with empty DataFrame
            
            ### gather general information ###
            station_id = station.get("_id",np.nan)
            # add id to dict
            stationdict["id"].append(station_id)

            ### gather positional information ###
            place = station.get("place",{})
            location = place.get("location",[np.nan,np.nan])
            longitude,latitude = location # get longitude and latitude
            altitude = place.get("altitude",np.nan)
            timezone = place.get("timezone",np.nan)
            # add position to dict
            stationdict["altitude"].append(altitude)
            stationdict["longitude"].append(longitude)
            stationdict["latitude"].append(latitude)
            stationdict["timezone"].append(timezone)

            ### gather measurement information ###
            # loop over all measurements
            measdict = {}
            for module_id, measure in station.get("measures",{}).items():
                types = measure.get("type",[])
                res   = measure.get("res",{}) # the time and values
                if not len(res) == 1: # something is wrong
                    # self.logger.warning("module '{}' has not exactly one time! " 
                    #     "Leaving it out.".format(module_id))
                    res = {np.nan:[np.nan] * len(types)}
                timestamp = list(res.keys())[0]
                measurements = res.get(timestamp)
                measdict.update( zip(types,measurements) )
                timedict = { "time_{}".format(t): int(timestamp) for t in types }
                measdict.update( timedict  )
                # self.logger.debug("measdict after update: {}".format(measdict))

            # add measurements to stationdict
            for key,val in measdict.items():
                try:    stationdict[key].append(val)
                except: stationdict[key] = [val]
                
            # fill the remaining values
            length = max([len(x) for x in stationdict.values()])
            for key,val in stationdict.items():
                while len(val) < length: val.append(np.nan)
                
            
        # create DataFrame
        df = DataFrame(stationdict)
        # convert times to datetime
        for col in df.columns:
            if col.startswith("time_"):
                df[col] = pd.to_datetime(df[col], unit="s", utc=True)

        # drop outliers if desired
        if only_inside:
            lat_ne = self.request.payload["lat_ne"]
            lon_ne = self.request.payload["lon_ne"]
            lat_sw = self.request.payload["lat_sw"]
            lon_sw = self.request.payload["lon_sw"]
            # TODO: What happens in other regions? Is outside still correct?
            outside = \
                np.logical_or( df["latitude"]  > lat_ne, 
                np.logical_or( df["latitude"]  < lat_sw,
                np.logical_or( df["longitude"] > lon_ne,
                               df["longitude"] < lon_sw )))
            # drop outliers
            df.drop(np.where(outside)[0],inplace=True)

        # reset index
        df.reset_index(inplace = True)
        # return the resulting DataFrame
        return df
    

class GetmeasureResponse(ApiResponse):
    """ Class that holds the responded data of a Getmeasure request
    """
    def dataframe(self):
        """ 
        Convert the response to a pandas.DataFrame

        Returns:
            pandas.DataFrame: The response converted to a DataFrame
        """
        body = self.response.get("body")
        if isinstance(body,list): # optimized # pragma: no cover
            if body: # only if there really is something
                raise NotImplementedError("converting 'optimized' Getmeasure " 
                    "response is not yet implemented")
            else:
                body = {} # fake to empty dict
        if isinstance(body,dict): # unoptimized
            try:
                types = self.request.payload.get("type").split(",") # the types
            except AttributeError: # pragma: no cover
                raise InvalidApiInputError("There is no sensible 'type' " 
                    "section in the request's payload. Strange...")
                    
            # start with empty measurement dict
            measdict = {"time":[]}
            # loop over all time-measurement pairs
            for timestamp, measurements in body.items():
                if not len(types) == len(measurements): # pragma: no cover
                    raise ApiResponseError("number of requested types does not " 
                        "match number of responded types")
                d = dict(zip(types, measurements))
                d.update({"time":int(timestamp)})
                # add measurements to measdict
                for key,val in d.items():
                    try:    measdict[key].append(val)
                    except: measdict[key] = [val]

        else: # bullshit # pragma: no cover
            raise ApiResponseError("'body' part of response does not " 
                "exist or is neither list not dict.")

        # create DataFrame
        df = DataFrame(measdict)
        # convert times to datetime
        for col in df.columns:
            if col.startswith("time"):
                df[col] = pd.to_datetime(df[col], unit="s", utc=True)
        # sort the data frame by time
        df = df.sort_values(by = "time")
        # index
        df.set_index("time",inplace=True)
        # return the resulting DataFrame
        return df


class GetstationsdataResponse(ApiResponse):
    """ Class that holds the responded data of a Getstationsdata request
    """
    def dataframe(self):
        """ 
        Convert the response to a pandas.DataFrame

        Returns:
            pandas.DataFrame: The response converted to a DataFrame
        """
        body = self.response.get("body")
        if not isinstance(body,dict): # optimized # pragma: no cover
            raise ApiResponseError("api response body is no dict. Strange...")
        devices = body.get("devices")
        if not isinstance(devices, list): # no list # pragma: no cover
            raise ApiResponseError("api response devices part is no list. " 
                "Strange...")
        if len(devices) > 1: # pragma: no cover
            raise ApiResponseError("api response devices list is longer than " 
                "one. Strange...")
        try:
            dashboard_data = devices[0]["dashboard_data"] # device
        except KeyError: # pragma: no cover
            raise ApiResponseError("api response devices list entry has no " 
                "dashboard_data")

        # get proper data
        data = {} # start with empty dict
        for key, val in dashboard_data.items():
            if type(val) == dict or type(val) == list: # pragma: no cover
                raise ApiResponseError("dashboard_data values are not scalars")
            data[key] = [val] # put it into a list
                
        # convert to data frame
        df = pd.DataFrame(data)

        # reset index
        df.reset_index(inplace = True)

        return df



#!/usr/bin/env python3
# System modules
import json
import datetime

# External modules

# Internal modules


# Variables
EMPTY_JSON = {}

# read json from a filename
def read_json_from_file(filename):
    """
    read json from a file given then filename
    args:
        filename (str): The path to the file to read
    returns:
        dict, empty dict if error occured during read
    """
    try: # open and read, return result
        with open(filename, "r") as f:
            return json.load(f)
    except: # didn't work, return empty dict
        return {}

# write json to file
def write_json_to_file(data, filename):
    """
    write data to json file
    args:
        filename (str): the json file to write the data to
        data (dict): the data to write to the file
    returns:
        True on success, False otherwise
    """
    try:
        jsonstr = json.dumps(data,sort_keys=True,indent=4) # try to convert
        # write to file
        with open(filename, "w") as f: f.write(jsonstr)
        return True
    except:
        return False

# get UNIX-timestamp from datetime object
def unix_timestamp(dt):
    """ 
    Calculate the UNIX-timestamp - i.e. the seconds since 01.01.1970 - from a
    given datetime object

    Args:
        dt (:any:`datetime.datetime`): a datetime object
    Returns:
        float : the UNIX-timestamp in seconds
    """
    return (dt - datetime.datetime(1970,1,1)).total_seconds()

# do nothing
def nothing(): # pragma: no cover
    pass



#!/usr/bin/env python3
import sys
import json
import re
import os
import logging
import signal
import time
import contextlib

from .. import utils
from . import requests
from .variables import *
from .errors import *

# constants
EMPTY_TIMES = { 
"read_tokens":0,
"write_tokens":0,
"update_tokens":0,
"change_tmpfile":0,
"request_tokens":0,
"refresh_tokens":0,
}

TOKEN_REGEX = re.compile("^[a-z0-9]+\|[a-z0-9]+$")


EMPTY_CREDENTIALS = {
    "password"     :"",
    "username"     :"",
    "client_id"    :"",
    "client_secret":""
}

EMPTY_TOKEN_REQUEST = {
    "grant_type": "password",
    "password"     :"",
    "username"     :"",
    "client_id"    :"",
    "client_secret":"",
}

EMPTY_TOKEN_REFRESH_REQUEST = {
    "grant_type": "refresh_token",
    "client_id"    :"",
    "client_secret":"",
    "refresh_token":"",
}

DEFAULT_EXPIRE_TIME = 10800

EMPTY_TOKENS = { "access_token" :"", "refresh_token":"" }

class Authentication(object):
    """
    Netatmo api Oauth2 authentication client

    Args:
        credentials (dict of str, optional): developer account credentials. 
            Required keys: username, password, cliend_id, client_secret
        tokens (dict of str, optional): the access and refresh token.
            Required keys: access_token, refresh_token
        tmpfile (str, optional): temporary file to use for tokens
    """
    def __init__(self, 
            credentials = EMPTY_CREDENTIALS,
            tokens      = EMPTY_TOKENS,
            tmpfile     = None
            ):
        self.logger.debug("init: setting credentials...")
        self.credentials = credentials
        self.logger.debug("init: credentials set.")
        self.logger.debug("init: setting tmpfile...")
        self.tmpfile     = tmpfile 
        self.logger.debug("init: tmpfile set.")

        # only set the (empty) tokens if you didn't already read them from file
        if self.last_token_action != "read_tokens" or tokens != EMPTY_TOKENS: 
            self.logger.debug("init: setting tokens...")
            self.tokens      = tokens
            self.logger.debug("init: tokens set.")



    ##################
    ### Properties ###
    ##################
    @property
    def logger(self):
        """ the logging.Logger used for logging.
        Defaults to logging.getLogger(__name__).
        """
        try: # try to return the internal property
            return self._logger 
        except AttributeError: # didn't work
            return logging.getLogger(__name__) # return default logger
    
    @logger.setter
    def logger(self, logger): # pragma: no cover
        assert isinstance(logger, logging.Logger), \
            "logger property has to be a logging.Logger"
        self._logger = logger

    @property
    def credentials(self):
        """ dict of str: The developer account credentials

        If you set this property, it is checked for valid content first.
        """
        try: # try to return the internal attribute
            return self._credentials
        except AttributeError: # pragma: no cover
            return EMPTY_CREDENTIALS

    @credentials.setter
    def credentials(self, value):
        assert type(value) is dict, "credentials need to be given as dict"
        # check if all needed credentials are there
        for key in EMPTY_CREDENTIALS.keys():
            assert key in value.keys(), \
                "credentials '{}' not defined in new credentials".format(key)
        # set new credentials
        self._credentials = value

    @property
    def tokens(self):
        """ dict of str: The OAuth2 tokens

        If you set this property, it is checked for valid content first.
        Also, automatic saving to the tmpfile occurs if necessary.
        """
        # if there are no tokens yet, set them to empty tokens
        if not hasattr(self,"_tokens"): self._tokens = EMPTY_TOKENS

        if self.token_getter_in_progress: # okay, we don't want recursion
            self.logger.debug("tokens getter: to prevent recursion, don't make "
                "sure tokens are up to date.")
        else:
            self.logger.debug("tokens getter: making sure " 
                "tokens are up to date.")
            with self.no_token_getter_recursion():
                self.make_sure_tokens_are_up_to_date()

        return self._tokens

    @tokens.setter
    def tokens(self, value):
        if self.check_tokens(value):
            # set the new values
            self._tokens = value

            # update the update time
            self.update_time("update_tokens")

            # do tmpfile IO
            self.tmpfile_io()
        else:
            self.logger.info("given tokens have invalid format."
                    "Using empty default.")
            self._tokens = EMPTY_TOKENS

    @property
    def token_getter_in_progress(self):
        """ This is an internal property to prevent recursion in the tokens
        getter. It is always boolean. When you set it, it is converted to bool.
        """
        if hasattr(self,"_token_getter_in_progress"):
            return bool(self._token_getter_in_progress)
        else:
            return False

    @token_getter_in_progress.setter
    def token_getter_in_progress(self, value):
        self._token_getter_in_progress = bool(value)

    @property
    def tmpfile(self):
        """ str: The temporary file for the tokens.

        If you set this property to something DIFFERENT than before, an attempt
        to read tokens from the tmpfile is made. If that fails, the current
        tokens are written to the tmpfile.
        """
        try:    return self._tmpfile
        except: return None # defaults to None

    @tmpfile.setter
    def tmpfile(self, value):
        # set the value
        oldtmpfile = self.tmpfile
        if oldtmpfile == value:
            self.logger.debug((
            "Request to set the tmpfile to the same value of '{}'."
            " Not doing anything.").format(value))
            return

        self._tmpfile = value
        self.logger.debug("The tmpfile was changed from '{}' to '{}'.".format(
            oldtmpfile, value))

        # update the tmpfile change time
        self.update_time("change_tmpfile")

        # do tmpfile IO
        self.tmpfile_io()

    ### property checkers ###
    @property
    def tokens_defined(self):
        """ check if the 'tokens' property contains non-empty tokens

        Returns:
            bool: True if tokens are complete, False otherwise 
        """
        with self.no_token_getter_recursion():
            res = True
            for key in EMPTY_TOKENS.keys():
                if not TOKEN_REGEX.match(self.tokens.get(key,"")):
                    res = False
        return res

    @property
    def credentials_defined(self):
        """ check if the 'credentials' property contains full credentials

        Returns:
            bool: True if credentials are complete, False otherwise 
        """
        res = True
        for key in EMPTY_CREDENTIALS.keys():
            if not self.credentials.get(key):
                res = False
        return res

    @property
    def last_request_time(self):
        """ [read only] The UNIX timestamp of the last new token request
        """
        last_request_time = \
            self.tokens.get("token_request_time",    # try saved time
            self.token_times.get("request_tokens",  # try internal time
            0 # use default
            ))
        return last_request_time

    @property
    def last_refresh_time(self):
        """ [read only] The UNIX timestamp of the last new token request
        """
        last_refresh_time = \
            self.tokens.get("token_refresh_time",    # try saved time
            self.token_times.get("refresh_tokens",  # try internal time
            0 # use default
            ))
        return last_refresh_time

    @property
    def expire_time(self):
        """ [read only] The token expire time 
        """
        expire_time = self.tokens.get("expire_in",   # try expire_in
                      self.tokens.get("expires_in",  # else, try expires_in
                      DEFAULT_EXPIRE_TIME            # else, use default
                      ))
        return expire_time

    @property
    def tokens_are_up_to_date(self):
        """ check if the tokens are up to date

        Returns:
            bool: True if the tokens are still valid, False if they need to be
                updated.
        """
        with self.no_token_getter_recursion():
            now = time.time() # current time
            if not self.tokens_defined: # no tokens, not up to date
                return False
            # if the validity time span has been expired
            if self.last_refresh_time + self.expire_time < now and \
               self.last_request_time + self.expire_time < now:
                return False # not up to date
            else: # not expired
                return True # up to date


    ### the specific times ###
    @property
    def token_times(self):
        """ Internal time counter for token-related actions
        
        Defaults to all times equal to 0.
        """
        try:    return self._token_times
        except: return EMPTY_TIMES

    @property
    def last_token_action(self):
        """ Determine the last token-related action from token_times property
        
        Returns the name of the most recent token-related action and None if
        no action has happened yet.
        """
        if self.token_times == EMPTY_TIMES:
            # if nothing was done, return None
            self.logger.debug("There was no token-related action yet.")
            return None
        else:
            # return the most recent action
            action = sorted(                # sort
                self.token_times,         # the token_times
                key=self.token_times.get, # according to the times
                reverse = True            # in reversed order (newest first)
                )[0]                      # and return the newest element
            self.logger.debug("The last token-related action was '{}'.".format(
                action))
            return action

    ### context managers ###
    @contextlib.contextmanager
    def no_token_getter_recursion(self):
        try:
            old_token_getter_in_progress = self.token_getter_in_progress
            self.logger.debug("recursion-preventer: setting variable to True")
            self.token_getter_in_progress = True
            yield
        finally:
            self.logger.debug(
                "recursion-preventer: resetting variable to {}".format(
                old_token_getter_in_progress))
            self.token_getter_in_progress = old_token_getter_in_progress


    ###############
    ### Methods ###
    ###############
    def check_tokens(self, tokens):
        """ Check given tokens for completeness and syntax

        Args:
            tokens (dict of str): The tokens to check

        Returns:
            bool: True if tokens are complete, False otherwise
        """
        # check type
        if not type(tokens) is dict: return False
        # check if all needed credentials are there
        for key in EMPTY_TOKENS.keys():
            if not key in tokens.keys(): return False
        return True

    def update_time(self, action):
        """ update the internal time for the token-related action 'action'

        Args:
            action (str): the token-related action to update the time for
        """
        self.logger.debug("updating the time for action '{}'".format(action))
        if not action in EMPTY_TIMES.keys(): # pragma: no cover
            raise KeyError
        # update the tmpfile change time
        times = self.token_times.copy()
        times.update({action:time.time()})
        self._token_times = times

    def tmpfile_io(self):
        """ Read or write the tmpfile dependant on what happend last
        Dependant on the last_token_action property, determine read or write
        the tmpfile.
        """
        actions = {
        "change_tmpfile"  : self.read_tokens_from_tmpfile,
        "update_tokens"   : self.write_tokens_to_tmpfile,
        "read_tokens"     : utils.nothing,
        "write_tokens"    : utils.nothing,
        "request_tokens"  : utils.nothing,
        "refresh_tokens"  : utils.nothing,
        }

        last_token_action = self.last_token_action
        # get the appropriate action
        action = actions.get(last_token_action, utils.nothing)
        self.logger.debug("The appropriate action is {}.".format(
            action.__name__))

        # do it!
        success = action()

        # if reading from tmpfile didn't work out, write it
        if action == self.read_tokens_from_tmpfile and not success:
            self.write_tokens_to_tmpfile()
            
    def read_tokens_from_tmpfile(self):
        """ Read tokens from the tmpfile
        The time for the 'read_tokens'-action is updated on success.

        Returns:
            bool: True if something was actually read from the file into the
                tokens property, False otherwise.
        """
        self.logger.debug("Try reading tokens from tmpfile '{}'....".format(
            self.tmpfile))
        # read
        tokens = utils.read_json_from_file(self.tmpfile)
        # check
        if self.check_tokens(tokens):
            self._tokens = tokens # set
            self.update_time("read_tokens") # update time
            self.logger.debug("tokens were read from tmpfile '{}'.".format(
                self.tmpfile))
            return True
        else:
            self.logger.debug("tokens from tmpfile '{}' are invalid.".format(
                self.tmpfile))
            return False
            

    def write_tokens_to_tmpfile(self):
        """ Write the current tokens to the tmpfile
        The time for the 'write_tokens'-action is updated on success.

        Returns:
            bool: True if something was actually written to file, False
                otherwise.  
        """
        with self.no_token_getter_recursion():
            self.logger.debug("Try writing tokens to tmpfile '{}'.".format(
                self.tmpfile))
            # try to write
            success = utils.write_json_to_file(self.tokens, self.tmpfile)
            if success: # check
                self.logger.debug("tokens written to tmpfile '{}'".format(
                    self.tmpfile))
                self.update_time("write_tokens") # update time
            else:
                self.logger.debug("could not write tokens to tmpfile '{}'".format(
                    self.tmpfile))
                
            return success


    def request_new_tokens(self):
        """ POST a token request to the api OAuth2 server to get NEW tokens
        """
        # check for completeness
        assert self.credentials_defined, \
            "Cannot make a token request with incomplete credentials"

        with self.no_token_getter_recursion():
            payload = EMPTY_TOKEN_REQUEST.copy() # a copy of an empty request
            payload.update(self.credentials) # update with credentials
            self.logger.debug("token request payload: {}".format(payload))

            ### create the request ###
            # the request
            tokenrequest = requests.TokenRequest(payload = payload)
            # the response
            tokenresponse = tokenrequest.response

            # check for errors
            error = tokenresponse.response.get("error")
            if error: # pragma: no cover
                raise API_ERRORS.get(error, ApiResponseError(error))

            tokens = self.tokens.copy() # a copy of current tokens
            tokens.update(tokenresponse.response) # update with new data
            tokens.update({"token_request_time":time.time()})
            self.tokens = tokens # set the updated tokens

            # update the time
            self.update_time("request_tokens")

    def refresh_current_tokens(self):
        """ POST a refresh request to the api OAuth2 server to refresh tokens
        """
        assert self.tokens_defined, \
            "Cannot refresh the tokens with incomplete tokens"

        with self.no_token_getter_recursion():
            ### create the payload ###
            payload = EMPTY_TOKEN_REFRESH_REQUEST.copy() # a copy of empty request
            # add refresh token
            for key in ("refresh_token",): # add refresh_token to payload
                payload[key] = self.tokens.get(key)
            for key in ("client_id", "client_secret"): # add credentials to payload
                payload[key] = self.credentials.get(key)
            self.logger.debug("token refresh request payload: {}".format(payload))

            ### create the request ###
            # the request
            tokenrequest = requests.TokenRequest(payload = payload)
            # the response
            tokenresponse = tokenrequest.response

            # check for errors
            error = tokenresponse.response.get("error")
            if error: 
                raise API_ERRORS.get(error, ApiResponseError(error))

            tokens = self.tokens.copy() # a copy of current tokens
            tokens.update(tokenresponse.response) # update with new data
            tokens.update({"refresh_request_time":time.time()})
            self.tokens = tokens # set the updated tokens

            # update the time
            self.update_time("refresh_tokens")

    def make_sure_tokens_are_up_to_date(self):
        # """ Make sure the tokens are up to date (if possible)
        # """
        self.logger.debug("I was told to make sure the tokens are up to date.")
        if self.tokens_are_up_to_date: # check if they are outdated
            self.logger.debug("tokens are already up to date! " 
                "No needed to update anything.")
        else:
            self.logger.debug("tokens are outdated! " 
                "Now let's check if we have credentials.")
            if self.credentials_defined:
                self.logger.debug("Yes, we have credentials and should be able" 
                    " to post requests.")
                if self.tokens_defined: # there are tokens already
                    self.logger.debug("We already have tokens, let's " 
                        "refresh them.")
                    self.refresh_current_tokens() # refresh them
                else: # no tokens yet
                    self.logger.debug("We don't have any tokens yet, let's " 
                        "request new ones.")
                    self.request_new_tokens() # request new
            else:
                self.logger.debug("credentials are not defined. I can't do " 
                    "anything to make sure the tokens are up to date...")
            
    def __repr__(self): # pragma: no cover
        """ python representation of this object
        """
        # self.logger.debug("__repr__ called")
        with self.no_token_getter_recursion():
            reprstring = ("{classname}(\n" 
                  "credentials = {credentials},\n" 
                  "tokens = {tokens},\n" 
                  "tmpfile = {tmpfile}\n"
                  ")").format(
                classname="{module}.{name}".format(
                    name=self.__class__.__name__,module=self.__class__.__module__),
                credentials = json.dumps(self.credentials,sort_keys=True,indent=8),
                tokens = json.dumps(self.tokens,sort_keys=True,indent=8),
                tmpfile = self.tmpfile if not isinstance(self.tmpfile,str) \
                    else '"{}"'.format(self.tmpfile)
                )
        return reprstring
        





# -*- coding: utf-8 -*-
# System modules

# External modules

# Internal modules

from . import authentication
from . import client
from . import errors
from . import variables
from . import requests
from . import responsetypes

__all__ = ['authentication']




#!/usr/bin/env python3
# module for api errors

##############
### Errors ###
##############
class ApiResponseError(BaseException):
    pass

class InvalidCredentialsError(ApiResponseError):
    pass

class InvalidApiInputError(BaseException):
    pass

class InvalidRegionError(InvalidApiInputError):
    def __init__(self):
        message = \
            ("'region' required keys: " 
            "lat_ne [-85;85], lat_sw [-85;85], " 
            "lon_ne [-180;180] and lon_sw [-180;180] " 
            "with lat_ne > lat_sw and lon_ne > lon_sw")
        super().__init__(message)

class InvalidRequiredDataError(InvalidApiInputError):
    def __init__(self):
        message = "'required_data' must be None or in {}".format(
            GETPUBLICDATA_ALLOWED_REQUIRED_DATA)
        super().__init__(message)


class InvalidApiRequestInputError(BaseException):
    pass

class InvalidPayloadError(InvalidApiRequestInputError):
    pass

API_ERRORS = {
"invalid_client":  InvalidCredentialsError("wrong credentials"),
"invalid_request": ApiResponseError("invalid request"),
"invalid_grant":   ApiResponseError("invalid grant"),
"Device not found": ApiResponseError("Device not found - Check Device ID " 
    "and permissions")
}





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# System modules
import os
import glob
from setuptools import setup, find_packages

__version__ = "0.2.10"

def read_file(filename):
    with open(filename) as f:
        return f.read()

# run setup
# take metadata from setup.cfg
setup( 
    name = "patatmo",
    description = "painless access to the netatmo weather api",
    long_description = read_file("README.rst"),
    author = "Yann Büchau",
    author_email = "yann.buechau@web.de",
    keywords = "netatmo, api",
    version = __version__,
    license = 'GPLv3',
    url = 'https://gitlab.com/nobodyinperson/python3-patatmo',
    classifiers = [
	'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
	'Programming Language :: Python :: 3.5',
	'Operating System :: OS Independent',
	'Topic :: Home Automation',
	'Topic :: Internet',
	'Topic :: Scientific/Engineering :: Atmospheric Science',
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Topic :: Utilities',
        ],
    test_suite = 'tests',
    tests_require = [ 'pandas', 'numpy' ],
    install_requires = [ 'pandas', 'numpy' ],
    packages = find_packages(exclude=['tests']),
    scripts = glob.glob("bin/netatmo-*"),
    )




#!/usr/bin/env python3
# module for global variables
import re

PLAIN_URLENCODED_HEADERS = {
    "Content-type": "application/x-www-form-urlencoded;charset=UTF-8",
    }

NETATMO_API_SERVER = "api.netatmo.com"

NETATMO_API_TOKEN_URL           = "/oauth2/token"
NETATMO_API_GETPUBLICDATA_URL   = "/api/getpublicdata"
NETATMO_API_GETMEASURE_URL      = "/api/getmeasure"
NETATMO_API_GETSTATIONSDATA_URL = "/api/getstationsdata"

GETPUBLICDATA_ALLOWED_REQUIRED_DATA = \
    ["rain","humidity","pressure","wind","temperature"]


GETMEASURE_ALLOWED_TYPES_UPTO_MAX = ["Temperature", "CO2", "Humidity", 
    "Pressure", "Noise", "Rain", "WindStrength", "WindAngle", "GustStrength",
    "GustAngle"]

GETMEASURE_ALLOWED_TYPES_UPTO_3HOURS = GETMEASURE_ALLOWED_TYPES_UPTO_MAX.copy()
GETMEASURE_ALLOWED_TYPES_UPTO_3HOURS.extend([
    "min_temp", "max_temp", "min_hum", "max_hum", "min_pressure",
    "max_pressure", "min_noise", "max_noise", "sum_rain", "date_max_gust"])

GETMEASURE_ALLOWED_TYPES_UPTO_1MONTH = \
GETMEASURE_ALLOWED_TYPES_UPTO_3HOURS.copy()
GETMEASURE_ALLOWED_TYPES_UPTO_1MONTH.extend([
    "min_temp", "max_temp", "min_hum", "max_hum", "min_pressure",
    "max_pressure", "min_noise", "max_noise", "sum_rain", "date_max_gust"])

GETMEASURE_ALLOWED_TYPES_BY_SCALE = {
"max" :    GETMEASURE_ALLOWED_TYPES_UPTO_MAX,
"30min" :  GETMEASURE_ALLOWED_TYPES_UPTO_3HOURS,
"1hour" :  GETMEASURE_ALLOWED_TYPES_UPTO_3HOURS,
"3hours" : GETMEASURE_ALLOWED_TYPES_UPTO_3HOURS,
"1day" :   GETMEASURE_ALLOWED_TYPES_UPTO_1MONTH,
"1week" :  GETMEASURE_ALLOWED_TYPES_UPTO_1MONTH,
"1month" : GETMEASURE_ALLOWED_TYPES_UPTO_1MONTH,
}

GETPUBLICDATA_REGION_BOUNDS = {
    "lat_ne": [-85,85],
    "lat_sw": [-85,85],
    "lon_ne": [-180,180],
    "lon_sw": [-180,180],
    }

MAC_ADDRESS_REGEX = re.compile(":".join(["[0-9a-f]{2}"] * 6))



# -*- coding: utf-8 -*-
# System modules
import os,sys
import argparse
import logging
import configparser

# External modules

# Internal modules
from ..api.client import NetatmoClient
from ..api.authentication import Authentication

def cli_argumentparser(*args, **kwargs): # pragma: no cover
    """ 
    Set up and return an :any:`argparse.ArgumentParser` with default arguments
    every patatmo cli script needs.

    Args:
        args (list, optional): further positional arguments handed to the 
            constructor
        kwargs (dict, optional): further keyword arguments handed to the 
            constructor

    Returns:
        argparse.ArgumentParser: the parser with default arguments
    """
    parser = argparse.ArgumentParser( *args, **kwargs )
    parser.add_argument("-p","--password", 
        help = "Netatmo developer account password. " 
            "Takes precedence over NETATMO_PASSWORD environment variable", 
        required = False)
    parser.add_argument("-u","--user", 
        help = "Netatmo developer account username. "
            "Takes precedence over NETATMO_USERNAME environment variable", 
        required = False)
    parser.add_argument("-i","--id", 
        help = "Netatmo app client id. "
            "Takes precedence over NETATMO_CLIENT_ID environment variable", 
        required = False)
    parser.add_argument("-s","--secret", 
        help = "Netatmo app client secret. "
            "Takes precedence over NETATMO_CLIENT_SECRET environment variable", 
        required = False)
    parser.add_argument("-o","--output", 
        help = "output file. Defaults to '-' which means STDOUT.", default = "-")
    parser.add_argument("-f","--tmpfile", help = "temporary authentication file",
        required = False)
    parser.add_argument("--noconfig", help = "ignore the user's personal " 
        "patatmo configuration folder completely", required = False, 
        default = False, action = "store_true")
    parser.add_argument("-v","--verbose", help = "verbose output",
        action="store_true", default = False)
    return parser


def get_credentials(arguments = argparse.Namespace()): # pragma: no cover
    """
    Try to determine the netatmo credentials in this order:

    1) command-line arguments
    2) environment variables
    3) user configuration folder

    Raise an error if it was not possible to determine all credentials.

    Args:
        arguments (argparse.Namespace, optional): the parsed command-line
            arguments.

    Returns:
        dict : the credentials ready to use by :any:`Authentication`.

    Raises:
        CredentialsError : if any of the credentials could not be determined
    """
    def get_credential(longname,argname,confname,envname):
        try: # try to get 
            value = getattr(arguments,argname)
            return(value.strip())
        except AttributeError: # user not given via command-line
            value = os.environ.get(envname)
            try: # check if value is a string
                return(value.strip())
            except AttributeError: # not given in environment
                try: # try to read from config file
                    config = get_configuration() # get user configuration
                    return(config["account"].get(confname).strip())
                except (KeyError, AttributeError): # not in configuration file
                    methods = []
                    methods.append("command-line argument '{}'".format(argname))
                    methods.append("environment variable '{}'".format(envname))
                    if not (arguments.noconfig \
                        if hasattr(arguments,"noconfig") else False):
                        methods.append("'{}' key in 'accounts' section " 
                            "in patatmo user configuration file '{}'".format(
                            confname,get_configuration_file()))
                    if not methods:
                        methods.append("nothing :-) You should not see this.")
                        
                    methodstr=" or ".join(
                        x for x in [", ".join(methods[0:-1]),methods[-1]] if x)
                    message = "Specify {} via {}".format(longname,methodstr)
                    raise CredentialsError(message)
                    
                
    needed_credentials = [ 
        ["Netatmo account username","user","username","NETATMO_USERNAME",
            "username"],
        ["Netatmo account password","password","password","NETATMO_PASSWORD",
            "password"],
        ["Netatmo app id","id","client_id","NETATMO_CLIENT_ID","client_id"],
        ["Netatmo app secret","secret","client_secret","NETATMO_CLIENT_SECRET",
            "client_secret"],
        ]

    credentials = {}
    for longname,argname,confname,envname,resname in needed_credentials:
        credentials[resname] = get_credential(longname=longname,
            argname=argname,envname=envname,confname=confname)

    return credentials

def get_client(arguments = argparse.Namespace()): # pragma: no cover
    """ 
    Set up a client for direct usage.

    Args:
        arguments (argparse.Namespace, optional): the parsed command-line
            arguments.

    Returns:
        NetatmoClient : a ready-to-use netatmo api client

    Raises:
        CredentialsError : if any of the credentials could not be determined
    """
    # read credentials
    try:
        credentials = get_credentials(arguments=arguments)
    except CredentialsError as e:
        logging.error(str(e))
        sys.exit(1)

    if not (arguments.noconfig if hasattr(arguments,"noconfig") else False):
        configdir = get_configuration_dir()
        make_sure_dir_exists(configdir)
        tmpfilename = \
            ".{}.json".format(credentials.get("client_id","auth").strip())
        tmpfile = os.path.join(configdir,tmpfilename)
    else:
        tmpfile = (arguments.tmpfile if hasattr(arguments,"tmpfile") else None)

    # set up a client
    client = NetatmoClient(
        authentication = Authentication(
            credentials = credentials,
            tmpfile = tmpfile,
            ),
        )

    return client

def get_output_file(arguments = argparse.Namespace()): # pragma: no cover
    """ 
    Return a file handler according to the 'output' option in the given
    arguments.

    Args:
        arguments (argparse.Namespace, optional): the parsed command-line
            arguments.

    Returns:
        filehandle : the filehandle to output to
    """
    try:
        output = arguments.output
    except AttributeError:
        output = "-"
    return sys.stdout if output == '-' or not output else open(output,"w")
    

def get_configuration_dir(): # pragma: no cover
    """ 
    Return the user's configuration directory

    Returns:
        str : path to the user's configueration directory
    """
    return os.path.expanduser("~/.patatmo")

def make_sure_dir_exists(directory): # pragma: no cover
    """ 
    Make sure the given directory exists

    Args:
        directory (str): the path of interest
    """
    if not os.path.exists(directory):
        os.makedirs(directory)

def get_configuration_file(): # pragma: no cover
    """
    Return the user's configuration file

    Returns:
        str : path to the user's configueration file
    """
    return os.path.join(get_configuration_dir(),"settings.conf")

def get_configuration(): # pragma: no cover
    """
    Return the user's patatmo configuration

    Returns:
        configparser.ConfigParser: the user's patatmo configuration
    """
    config = configparser.ConfigParser()
    config.read( get_configuration_file() )
    return config
    
class CredentialsError(ValueError): # pragma: no cover
    pass
